@@ -1,42 +1,184 @@
 import asyncio
 import collections
 import warnings
+from types import TracebackType
+from typing import (
+    Any,
+    Awaitable,
+    Callable,
+    Deque,
+    Generator,
+    Optional,
+    Set,
+    Type,
+)
 
 import async_timeout
 import psycopg2.extensions
 
-from .connection import TIMEOUT, connect
-from .utils import (
-    _PoolAcquireContextManager,
-    _PoolConnectionContextManager,
-    _PoolContextManager,
-    _PoolCursorContextManager,
-    create_completed_future,
-    get_running_loop,
-)
+from .connection import TIMEOUT, Connection, Cursor, connect
+from .utils import _ContextManager, create_completed_future, get_running_loop
 
 
-def create_pool(dsn=None, *, minsize=1, maxsize=10,
-                timeout=TIMEOUT, pool_recycle=-1,
-                enable_json=True, enable_hstore=True, enable_uuid=True,
-                echo=False, on_connect=None,
-                **kwargs):
+def create_pool(
+    dsn: Optional[str] = None,
+    *,
+    minsize: int = 1,
+    maxsize: int = 10,
+    timeout: float = TIMEOUT,
+    pool_recycle: float = -1.0,
+    enable_json: bool = True,
+    enable_hstore: bool = True,
+    enable_uuid: bool = True,
+    echo: bool = False,
+    on_connect: Optional[Callable[[Connection], Awaitable[None]]] = None,
+    **kwargs: Any,
+) -> _ContextManager["Pool"]:
     coro = Pool.from_pool_fill(
-        dsn, minsize, maxsize, timeout,
-        enable_json=enable_json, enable_hstore=enable_hstore,
-        enable_uuid=enable_uuid, echo=echo, on_connect=on_connect,
-        pool_recycle=pool_recycle, **kwargs
+        dsn,
+        minsize,
+        maxsize,
+        timeout,
+        enable_json=enable_json,
+        enable_hstore=enable_hstore,
+        enable_uuid=enable_uuid,
+        echo=echo,
+        on_connect=on_connect,
+        pool_recycle=pool_recycle,
+        **kwargs,
     )
-
-    return _PoolContextManager(coro)
+    return _ContextManager[Pool](coro, _destroy_pool)
 
 
-class Pool(asyncio.AbstractServer):
+async def _destroy_pool(pool: "Pool") -> None:
+    pool.close()
+    await pool.wait_closed()
+
+
+class _PoolConnectionContextManager:
+    """Context manager.
+
+    This enables the following idiom for acquiring and releasing a
+    connection around a block:
+
+        async with pool as conn:
+            cur = await conn.cursor()
+
+    while failing loudly when accidentally using:
+
+        with pool:
+            <block>
+    """
+
+    __slots__ = ("_pool", "_conn")
+
+    def __init__(self, pool: "Pool", conn: Connection):
+        self._pool: Optional[Pool] = pool
+        self._conn: Optional[Connection] = conn
+
+    def __enter__(self) -> Connection:
+        assert self._conn
+        return self._conn
+
+    def __exit__(
+        self,
+        exc_type: Optional[Type[BaseException]],
+        exc: Optional[BaseException],
+        tb: Optional[TracebackType],
+    ) -> None:
+        if self._pool is None or self._conn is None:
+            return
+        try:
+            self._pool.release(self._conn)
+        finally:
+            self._pool = None
+            self._conn = None
+
+    async def __aenter__(self) -> Connection:
+        assert self._conn
+        return self._conn
+
+    async def __aexit__(
+        self,
+        exc_type: Optional[Type[BaseException]],
+        exc: Optional[BaseException],
+        tb: Optional[TracebackType],
+    ) -> None:
+        if self._pool is None or self._conn is None:
+            return
+        try:
+            await self._pool.release(self._conn)
+        finally:
+            self._pool = None
+            self._conn = None
+
+
+class _PoolCursorContextManager:
+    """Context manager.
+
+    This enables the following idiom for acquiring and releasing a
+    cursor around a block:
+
+        async with pool.cursor() as cur:
+            await cur.execute("SELECT 1")
+
+    while failing loudly when accidentally using:
+
+        with pool:
+            <block>
+    """
+
+    __slots__ = ("_pool", "_conn", "_cursor")
+
+    def __init__(self, pool: "Pool", conn: Connection, cursor: Cursor):
+        self._pool = pool
+        self._conn = conn
+        self._cursor = cursor
+
+    def __enter__(self) -> Cursor:
+        return self._cursor
+
+    def __exit__(
+        self,
+        exc_type: Optional[Type[BaseException]],
+        exc: Optional[BaseException],
+        tb: Optional[TracebackType],
+    ) -> None:
+        try:
+            self._cursor.close()
+        except psycopg2.ProgrammingError:
+            # seen instances where the cursor fails to close:
+            #   https://github.com/aio-libs/aiopg/issues/364
+            # We close it here so we don't return a bad connection to the pool
+            self._conn.close()
+            raise
+        finally:
+            try:
+                self._pool.release(self._conn)
+            finally:
+                self._pool = None  # type: ignore
+                self._conn = None  # type: ignore
+                self._cursor = None  # type: ignore
+
+
+class Pool:
     """Connection pool"""
 
-    def __init__(self, dsn, minsize, maxsize, timeout, *,
-                 enable_json, enable_hstore, enable_uuid, echo,
-                 on_connect, pool_recycle, **kwargs):
+    def __init__(
+        self,
+        dsn: str,
+        minsize: int,
+        maxsize: int,
+        timeout: float,
+        *,
+        enable_json: bool,
+        enable_hstore: bool,
+        enable_uuid: bool,
+        echo: bool,
+        on_connect: Optional[Callable[[Connection], Awaitable[None]]],
+        pool_recycle: float,
+        **kwargs: Any,
+    ):
         if minsize < 0:
             raise ValueError("minsize should be zero or greater")
         if maxsize < minsize and maxsize != 0:
@@ -53,38 +195,40 @@ class Pool(asyncio.AbstractServer):
         self._on_connect = on_connect
         self._conn_kwargs = kwargs
         self._acquiring = 0
-        self._free = collections.deque(maxlen=maxsize or None)
+        self._free: Deque[Connection] = collections.deque(
+            maxlen=maxsize or None
+        )
         self._cond = asyncio.Condition()
-        self._used = set()
-        self._terminated = set()
+        self._used: Set[Connection] = set()
+        self._terminated: Set[Connection] = set()
         self._closing = False
         self._closed = False
 
     @property
-    def echo(self):
+    def echo(self) -> bool:
         return self._echo
 
     @property
-    def minsize(self):
+    def minsize(self) -> int:
         return self._minsize
 
     @property
-    def maxsize(self):
+    def maxsize(self) -> Optional[int]:
         return self._free.maxlen
 
     @property
-    def size(self):
+    def size(self) -> int:
         return self.freesize + len(self._used) + self._acquiring
 
     @property
-    def freesize(self):
+    def freesize(self) -> int:
         return len(self._free)
 
     @property
-    def timeout(self):
+    def timeout(self) -> float:
         return self._timeout
 
-    async def clear(self):
+    async def clear(self) -> None:
         """Close all free connections in pool."""
         async with self._cond:
             while self._free:
@@ -93,10 +237,10 @@ class Pool(asyncio.AbstractServer):
             self._cond.notify()
 
     @property
-    def closed(self):
+    def closed(self) -> bool:
         return self._closed
 
-    def close(self):
+    def close(self) -> None:
         """Close pool.
 
         Mark all pool connections to be closed on getting back to pool.
@@ -106,7 +250,7 @@ class Pool(asyncio.AbstractServer):
             return
         self._closing = True
 
-    def terminate(self):
+    def terminate(self) -> None:
         """Terminate pool.
 
         Close pool with instantly closing all acquired connections also.
@@ -120,18 +264,19 @@ class Pool(asyncio.AbstractServer):
 
         self._used.clear()
 
-    async def wait_closed(self):
+    async def wait_closed(self) -> None:
         """Wait for closing all pool's connections."""
 
         if self._closed:
             return
         if not self._closing:
-            raise RuntimeError(".wait_closed() should be called "
-                               "after .close()")
+            raise RuntimeError(
+                ".wait_closed() should be called " "after .close()"
+            )
 
         while self._free:
             conn = self._free.popleft()
-            conn.close()
+            await conn.close()
 
         async with self._cond:
             while self.size > self.freesize:
@@ -139,13 +284,13 @@ class Pool(asyncio.AbstractServer):
 
         self._closed = True
 
-    def acquire(self):
+    def acquire(self) -> _ContextManager[Connection]:
         """Acquire free connection from the pool."""
         coro = self._acquire()
-        return _PoolAcquireContextManager(coro, self)
+        return _ContextManager[Connection](coro, self.release)
 
     @classmethod
-    async def from_pool_fill(cls, *args, **kwargs):
+    async def from_pool_fill(cls, *args: Any, **kwargs: Any) -> "Pool":
         """constructor for filling the free pool with connections,
         the number is controlled by the minsize parameter
         """
@@ -156,7 +301,7 @@ class Pool(asyncio.AbstractServer):
 
         return self
 
-    async def _acquire(self):
+    async def _acquire(self) -> Connection:
         if self._closing:
             raise RuntimeError("Cannot acquire connection after closing pool")
         async with async_timeout.timeout(self._timeout), self._cond:
@@ -171,7 +316,7 @@ class Pool(asyncio.AbstractServer):
                 else:
                     await self._cond.wait()
 
-    async def _fill_free_pool(self, override_min):
+    async def _fill_free_pool(self, override_min: bool) -> None:
         # iterate over free connections and remove timeouted ones
         n, free = 0, len(self._free)
         while n < free:
@@ -179,7 +324,7 @@ class Pool(asyncio.AbstractServer):
             if conn.closed:
                 self._free.pop()
             elif -1 < self._recycle < self._loop.time() - conn.last_usage:
-                conn.close()
+                await conn.close()
                 self._free.pop()
             else:
                 self._free.rotate()
@@ -189,12 +334,14 @@ class Pool(asyncio.AbstractServer):
             self._acquiring += 1
             try:
                 conn = await connect(
-                    self._dsn, timeout=self._timeout,
+                    self._dsn,
+                    timeout=self._timeout,
                     enable_json=self._enable_json,
                     enable_hstore=self._enable_hstore,
                     enable_uuid=self._enable_uuid,
                     echo=self._echo,
-                    **self._conn_kwargs)
+                    **self._conn_kwargs,
+                )
                 if self._on_connect is not None:
                     await self._on_connect(conn)
                 # raise exception if pool is closing
@@ -205,16 +352,18 @@ class Pool(asyncio.AbstractServer):
         if self._free:
             return
 
-        if override_min and self.size < self.maxsize:
+        if override_min and self.size < (self.maxsize or 0):
             self._acquiring += 1
             try:
                 conn = await connect(
-                    self._dsn, timeout=self._timeout,
+                    self._dsn,
+                    timeout=self._timeout,
                     enable_json=self._enable_json,
                     enable_hstore=self._enable_hstore,
                     enable_uuid=self._enable_uuid,
                     echo=self._echo,
-                    **self._conn_kwargs)
+                    **self._conn_kwargs,
+                )
                 if self._on_connect is not None:
                     await self._on_connect(conn)
                 # raise exception if pool is closing
@@ -223,13 +372,12 @@ class Pool(asyncio.AbstractServer):
             finally:
                 self._acquiring -= 1
 
-    async def _wakeup(self):
+    async def _wakeup(self) -> None:
         async with self._cond:
             self._cond.notify()
 
-    def release(self, conn):
-        """Release free connection back to the connection pool.
-        """
+    def release(self, conn: Connection) -> "asyncio.Future[None]":
+        """Release free connection back to the connection pool."""
         future = create_completed_future(self._loop)
         if conn in self._terminated:
             assert conn.closed, conn
@@ -244,7 +392,7 @@ class Pool(asyncio.AbstractServer):
             warnings.warn(
                 f"Invalid transaction status on "
                 f"released connection: {transaction_status}",
-                ResourceWarning
+                ResourceWarning,
             )
             conn.close()
             return future
@@ -254,15 +402,26 @@ class Pool(asyncio.AbstractServer):
             self._free.append(conn)
         return asyncio.ensure_future(self._wakeup(), loop=self._loop)
 
-    async def cursor(self, name=None, cursor_factory=None,
-                     scrollable=None, withhold=False, *, timeout=None):
+    async def cursor(
+        self,
+        name: Optional[str] = None,
+        cursor_factory: Any = None,
+        scrollable: Optional[bool] = None,
+        withhold: bool = False,
+        *,
+        timeout: Optional[float] = None,
+    ) -> _PoolCursorContextManager:
         conn = await self.acquire()
-        cur = await conn.cursor(name=name, cursor_factory=cursor_factory,
-                                scrollable=scrollable, withhold=withhold,
-                                timeout=timeout)
-        return _PoolCursorContextManager(self, conn, cur)
+        cursor = await conn.cursor(
+            name=name,
+            cursor_factory=cursor_factory,
+            scrollable=scrollable,
+            withhold=withhold,
+            timeout=timeout,
+        )
+        return _PoolCursorContextManager(self, conn, cursor)
 
-    def __await__(self):
+    def __await__(self) -> Generator[Any, Any, _PoolConnectionContextManager]:
         # This is not a coroutine.  It is meant to enable the idiom:
         #
         #     with (await pool) as conn:
@@ -278,23 +437,34 @@ class Pool(asyncio.AbstractServer):
         conn = yield from self._acquire().__await__()
         return _PoolConnectionContextManager(self, conn)
 
-    def __enter__(self):
+    def __enter__(self) -> "Pool":
         raise RuntimeError(
-            '"await" should be used as context manager expression')
+            '"await" should be used as context manager expression'
+        )
 
-    def __exit__(self, *args):
+    def __exit__(
+        self,
+        exc_type: Optional[Type[BaseException]],
+        exc: Optional[BaseException],
+        tb: Optional[TracebackType],
+    ) -> None:
         # This must exist because __enter__ exists, even though that
         # always raises; that's how the with-statement works.
         pass  # pragma: nocover
 
-    async def __aenter__(self):
+    async def __aenter__(self) -> "Pool":
         return self
 
-    async def __aexit__(self, exc_type, exc_val, exc_tb):
+    async def __aexit__(
+        self,
+        exc_type: Optional[Type[BaseException]],
+        exc: Optional[BaseException],
+        tb: Optional[TracebackType],
+    ) -> None:
         self.close()
         await self.wait_closed()
 
-    def __del__(self):
+    def __del__(self) -> None:
         try:
             self._free
         except AttributeError:
@@ -306,5 +476,5 @@ class Pool(asyncio.AbstractServer):
                 conn.close()
                 left += 1
             warnings.warn(
-                f"Unclosed {left} connections in {self!r}",
-                ResourceWarning)
+                f"Unclosed {left} connections in {self!r}", ResourceWarning
+            )
